import numpy as np
# import pyproj
import math


class GTOC(object):
    """
    Class that implements the Ground Track Oblique Cassini projection.
    The GTOC projection is uniquely defined by 5 parameters

    """
    def __init__(self, lon_0=None, lat_0=None, azRot=None, thetaEq=None, scale=None):
        """
        !@brief Init Method

        @param lon_0:   Longitude of the centre of the projection
        @param lat_0:   Latitude of the centre of the projection
        @param azRot:   Azimuth Rotation Angle
        @param thetaEq: Theta rotation parameter
        @param scale:   Scaling Factor
        """
        self.__lon_0 = np.deg2rad(lon_0)            ## Longitude of the centre of the projection
        self.__lat_0 = np.deg2rad(lat_0)            ## Latitude of the centre of the projection
        self.__azRot = np.deg2rad(azRot)            ## Azimuth Rotation Angle
        self.__thetaEq = np.deg2rad(thetaEq)        ## Theta rotation parameter
        self.__scale = scale                        ## Scaling Factor
        self.__EarthRadius = scale                  ## Earth Radius

        self.__A__ = 6378137.0
        self.__B__ = 6356752.3
        self.__E2__ = (1.- pow(self.__B__,2)/pow(self.__A__,2))
        self.__E2P__ = (pow(self.__A__,2)/pow(self.__B__,2) - 1.)

        self.__R_E2M__ = None

        self.__compute_scale()
        #Matrix R Initialisation
        self.__R_Earth2Map__()
        self.__R_Map2Earth()


    def Earth2Map(self, lon=None, lat=None):

        """
        !@brief:    Method to convert coordinate in WGS84 Lat/lon into the GTOC projection

        @param lon: input longitude array
        @param lat: input latitude array

        @return: projected X/Y coordinates in the GTOC reference system
        """

        # Conversion to Spherical System
        rho = self.__EarthRadius
        az = np.radians(lon)
        el = self.__ell2sph__(np.radians(lat))

        # Convert Spherical to Cartesian
        x_in, y_in, z_in = self.__sph2cart__(rho, az, el)

        # Perform Rotation
        x_out, y_out, z_out = np.matmul(self.__R_E2M__, [x_in, y_in, z_in])

        # Convert Cartesian to Spherical
        rho, az, el = self.__cart2sph__(x_out,y_out, z_out )

        # Apply Scale and Offset factors
        x = -self.__scale * (el+self.__thetaEq)
        y = (self.__scale * math.cos(self.__thetaEq) )* az

        return x,y


    def Map2Earth(self, x=None, y=None):
        """
        !@brief:    Method to convert coordinate in X/Y GTOC coordinates back into the Lat/Lon WGS84 reference system

        @param x: Input X coordinate in GTOC projection
        @param y: Input Y coordinate in GTOC projection

        @return: Lon/Lat Coordinates
        """
        ## Conversion to sperical coordinates
        rho = self.__EarthRadius
        az = y /(math.cos(self.__thetaEq) * self.__scale)
        el = -x / self.__scale - self.__thetaEq

        # Conversion to Cartesian
        x_in, y_in, z_in = self.__sph2cart__(rho, az, el)

        # Perform Rotation
        x_out, y_out, z_out = np.matmul(self.__R_M2E__, [x_in, y_in, z_in])

        # Conversion to Spherical Coordinates
        rho, az, el = self.__cart2sph__(x_out, y_out, z_out)

        lon = np.rad2deg(az)
        lat = np.rad2deg(self.__sph2ell__(el))

        return lon, lat


    def test_conversion(self, radius, lon, lat):

        x,y, z = self.__sph2cart__(radius, lon, lat)
        print(x,y,z)
        rad, lon, lat = self.__cart2sph__(x,y,z)
        print(rad, np.rad2deg(lon), np.rad2deg(lat))


    def __ell2sph__(self, ang):

        """
        !@brief:   Private Method to perform the conversion from elliptical to spherical angle

        @param ang: input angle
        @return: converted angle
        """
        return np.arctan( pow(self.__B__,2)/pow(self.__A__,2)*np.tan(ang))

    def __sph2ell__(self, elev):
        """
        !@brief:  Private method to perform the conversion from spherical to elliptical angle

        @param elev: Input elevation angle
        @return:  converted angle
        """

        return np.arctan( pow(self.__A__,2)/pow(self.__B__,2)*np.tan(elev))

    def __sph2cart__(self, rho, az, el):

        """
        !@brief     Private Method to convert spherical to cartesian coordinates

        @param rho: Rho Input distance
        @param az:  Azimuth Angle
        @param el: Elevation Angle

        @return: X,Y,Z cartesian coordinate
        """

        cosTheta = np.cos(el)

        x = rho * cosTheta * np.cos(az)
        y = rho * cosTheta * np.sin(az)
        z = rho * np.sin(el)

        return x,y,z

    def __cart2sph__(self, x, y, z):
        """
        !@brief:    Private Method to convert cartesian to spherical coordinates

        @param x: X input coordinate
        @param y: Y input coordinate
        @param z: X input coordinate

        @return: rho, az, el  - Spherical Coordinates
        """
        rho = np.sqrt(np.power(x,2) + np.power(y,2) + np.power(z,2))
        el = np.arcsin(z/rho)
        az = np.arctan2(y,x)

        return rho, az, el

    def __Rx__(self, alpha):

        return np.array([ [1. ,  0., 0.],
                          [0. , math.cos(alpha) , -math.sin(alpha)],
                          [0. , math.sin(alpha) , math.cos(alpha)]])

    def __Ry__(self, alpha):
        return np.array([[math.cos(alpha), 0., math.sin(alpha)],
                         [0., 1., 0.],
                         [-math.sin(alpha), 0., math.cos(alpha)]])

    def __Rz__(self, alpha):
        return np.array([[math.cos(alpha), -math.sin(alpha),0.],
                         [math.sin(alpha), math.cos(alpha), 0.],
                         [0., 0., 1.]])

    def __R_Earth2Map__(self):

        ## Converting Geotedic to Spheric

        elRef = self.__ell2sph__(self.__lat_0)
        azRef = self.__lon_0


        tmp1 = np.matmul(self.__Ry__(self.__thetaEq), self.__Rx__(-self.__azRot))
        tmp2 = np.matmul(self.__Ry__(self.__ell2sph__(self.__lat_0)) ,
                         self.__Rz__(-self.__lon_0))

        self.__R_E2M__ = np.matmul(tmp1,tmp2)

    def __R_Map2Earth(self):

        if self.__R_E2M__ is None:
            self.__R_Earth2Map__()
        else:
            self.__R_M2E__ = np.transpose(self.__R_E2M__)


    def __compute_scale(self):

        if self.__scale is  None:
            CLat = math.cos(self.__lat_0)
            Clat2 = pow(CLat,2)
            SLat2 = 1. - Clat2
            A2 = pow(self.__A__,2)
            B2 = pow(self.__B__,2)
            self.__scale = math.sqrt( (pow(A2,2)*Clat2 + pow(B2,2)*SLat2) /( A2*Clat2 + B2*SLat2) )
            self.__EarthRadius = self.__scale